/*
 * Copyright (c) 2005-2010 KOM – Multimedia Communications Lab
 *
 * This file is part of PeerfactSim.KOM.
 * 
 * PeerfactSim.KOM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * 
 * PeerfactSim.KOM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with PeerfactSim.KOM.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package de.tud.kom.p2psim.impl.topology.movement.aerial;


import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.Map;
import org.apache.commons.math3.geometry.euclidean.twod.Vector2D;
import de.tud.kom.p2psim.api.topology.movement.UAVMovementModel;
import de.tud.kom.p2psim.impl.energy.components.ActuatorComponent;
import de.tud.kom.p2psim.impl.energy.components.StatelessActuatorComponent;
import de.tud.kom.p2psim.impl.topology.component.UAVTopologyComponent;
import de.tud.kom.p2psim.impl.topology.util.PositionVector;
import de.tudarmstadt.maki.simonstrator.api.Time;
import de.tudarmstadt.maki.simonstrator.api.uavsupport.callbacks.ReachedLocationCallback;

/**
 * Simplified thrust-based local movement model based on the Intel Aero UAV. 
 * 
 * The movement logic uses only straight forward movement with the maximum speed available. 
 * 
 * TODO Acceleration
 * TODO Movement model for plane-like UAVs 
 * 
 * @author Julian Zobel
 * @version 1.0, 11.09.2018
 */
public class MulticopterMovement implements UAVMovementModel  {

	private UAVTopologyComponent topologyComponent;
	
	private double currentAngleOfAttack;
	private double currentSpeed;
	
	private double targetSpeed;	
	
	private LinkedList<PositionVector> route = new LinkedList<>();
	private Map<PositionVector, ReachedLocationCallback> locationCallbacks = new LinkedHashMap<>();  // TODO callback interface
		
	private StatelessActuatorComponent motor;
	
	private double mass = 1.465; // kg
	private final double airdensity = 1.2045; // kg/m^3
	private final double gravity = 9.807; // m/s^2
	private double A_top = 0.245; // m^2
	private double A_front = 0.1; // m^2
	private double dragCoefficient = 0.7;
	private double maxPitchAngle = Math.toRadians(60); // 45° max angle
	private double descentVelocityMax = 5; // m/s
	private double maxTurnAngle = Math.toRadians(90); // 90° per second turn angle
	
	
	public MulticopterMovement(UAVTopologyComponent topologyComponent) {
		this.topologyComponent = topologyComponent;			
	}
	
	boolean first = true;
	
	@Override
	public void move(long timeBetweenMovementOperations) {
				
		if(motor.isOn() && !route.isEmpty()) {
			PositionVector position = new PositionVector(topologyComponent.getRealPosition());			
						
			PositionVector target = route.getFirst();
			
			double distanceToTargetPosition = position.distanceTo(target);
			
			
			// If target point is reached within a 1 meter margin, we remove that point from the list 
			if(distanceToTargetPosition < 0.1 || distanceToTargetPosition < currentSpeed)
			{						
				target = route.removeFirst();
								
				if(route.isEmpty()) {	
					
					// go to hover mode					
					topologyComponent.updateCurrentLocation(target.clone()); 
					
					currentSpeed = 0;					
					motor.requestThrust(hoverThrustRequired());
					
					PositionVector direction = topologyComponent.getCurrentDirection().clone();
					direction.setEntry(2, 0);
					topologyComponent.updateCurrentDirection(direction);
					
					locationReached(topologyComponent.getCurrentLocation());
					return;
				}
				else {
					
					// get to speed	
					if(targetSpeed > 0 && targetSpeed < horizontalMaxVelocity()) {
						motor.requestThrust(estimateRequiredThrust(targetSpeed));
						currentSpeed = targetSpeed;
					}
					else {
						motor.requestThrust(horizontalMaxVelocityRequiredTotalThrust());
						currentSpeed = horizontalMaxVelocity();
					}							
			
					long timeUntilReachedLocation = (long) (distanceToTargetPosition / currentSpeed) * Time.SECOND;
					
					target = route.getFirst();					
					PositionVector directionToTarget = new PositionVector(target);
					directionToTarget.subtract(position);
					
					double timefactor = timeUntilReachedLocation / Time.SECOND;
					
					directionToTarget.normalize();
					topologyComponent.updateCurrentDirection(directionToTarget.clone());
					directionToTarget.multiplyScalar(currentSpeed * timefactor);
					
					PositionVector newPosition = new PositionVector(position);
					newPosition.add(directionToTarget);		
					
					topologyComponent.updateCurrentLocation(newPosition);
					
					if(timeUntilReachedLocation < timeBetweenMovementOperations) {
						this.move(timeBetweenMovementOperations - timeUntilReachedLocation);					
					}
					
				}			
				
			}
			else {
				double timefactor = timeBetweenMovementOperations / Time.SECOND;
				
				// get to speed	
				if(targetSpeed > 0 && targetSpeed < horizontalMaxVelocity()) {
					motor.requestThrust(estimateRequiredThrust(targetSpeed));
					currentSpeed = targetSpeed;
				}
				else {
					motor.requestThrust(horizontalMaxVelocityRequiredTotalThrust());
					currentSpeed = horizontalMaxVelocity();
				}				
				
				PositionVector directionToTarget = new PositionVector(target);
				directionToTarget.subtract(position);
				
				directionToTarget.normalize();
				
				if(directionToTarget.getX() != 0 || directionToTarget.getY() != 0) {
					topologyComponent.updateCurrentDirection(directionToTarget.clone());
				}				
				
				directionToTarget.multiplyScalar(currentSpeed * timefactor);

				PositionVector newPosition = new PositionVector(position);
				newPosition.add(directionToTarget);		
				
				topologyComponent.updateCurrentLocation(newPosition);
			}
			
		}
		else if(motor.isOn()) {
			
			if(currentSpeed != 0) {
				throw new UnsupportedOperationException("no route but speed not 0?");
			}
			
			PositionVector position = new PositionVector(topologyComponent.getRealPosition());	
			
			if(position.getAltitude() == 0) {
				motor.requestThrust(0);
			}
			else {
				motor.requestThrust(hoverThrustRequired());
			}
		}
	}
	
	/*
	 * 
	 */
	
	public double verticalDescentMaxThrust() {		
		// m * g - 0.5 * p * C * A * v^2
		return hoverThrustRequired() - 0.5 * bodyDrag(0, new PositionVector(0,0,1)) * descentVelocityMax * descentVelocityMax;
	}
	
	public double verticalAscentMaxAcceleration() {
		return (motor.getMaxThrust() - hoverThrustRequired()) / mass;
	}
	
	public double verticalAscentMaxVelocity() {		
		double maxThrust = motor.getMaxThrust();		
		return Math.sqrt(2.0 * (maxThrust - hoverThrustRequired()) / bodyDrag(0, new PositionVector(0,0,1)));		
	}
	
	
	public double hoverThrustRequired()  {		
		return mass * gravity;		
	}
	
	
	public double horizontalMaxVelocity() {
		
		double horizontalThrust = horizontalComponentMaxThrust();
				
		double maxVelocity = Math.sqrt( (2.0 * horizontalThrust) / bodyDrag(maxPitchAngle, new PositionVector(1,0,0)));
				
		return maxVelocity;
	}
	
	public double horizontalComponentMaxThrust() {
		
		// hoverthrust / cos => amount of thrust in horizonal direction with °angle		
		double stableAltitudeMaximumTotalThrust = horizontalMaxVelocityRequiredTotalThrust();
	
		// fraction of total thrust in horizonal (forward) direction with °angle
		double maximumHorizontalThrustStableAltitude = stableAltitudeMaximumTotalThrust * Math.sin(maxPitchAngle);
		
		return maximumHorizontalThrustStableAltitude;
	}

	public double horizontalMaxVelocityRequiredTotalThrust() {
		return hoverThrustRequired() / Math.cos(maxPitchAngle);
	}
	
	public double bodyDrag(double angleRadians, PositionVector direction) {
		return airdensity * dragCoefficient * areaExposedToDrag(angleRadians, direction);
	}
	
	public double areaExposedToDrag(double angleRadians, PositionVector direction) {
			
		Vector2D v = new Vector2D(Math.abs(direction.getX()) + Math.abs(direction.getY()), Math.abs(direction.getZ()));
		v = v.normalize();
		
		double areaExposedFront = v.getX() * (Math.sin(angleRadians) * A_top + Math.cos(angleRadians) * A_front );
		double areaExposedTop = v.getY() * (Math.cos(angleRadians) * A_top + Math.sin(angleRadians) * A_front);
		
		return areaExposedFront + areaExposedTop;
	}
	
	/*
	 * F_drag [N] = 0.5 * p * C_drag * A * v^2
	 */
	public double currentDrag() {
		return 0.5 * bodyDrag(currentAngleOfAttack, topologyComponent.getCurrentDirection()) * currentSpeed * currentSpeed;
	}
	
	/**
	 * Calculate the drag induced on the UAV with a given velocity and an angle of attack (in radians) moving forward horizontally.
	 * 
	 * @param velocity
	 * @param angleInRadians
	 * @return
	 */
	private double forwardDrag(double velocity, double angleInRadians) {
		return 0.5 * bodyDrag(angleInRadians, new PositionVector(1,0,0)) * velocity * velocity;
	}
	
	
	/*
	 *
	 */
	
	@Override
	public void setMotorControl(ActuatorComponent motor) {
		this.motor = (StatelessActuatorComponent) motor;
	}

	

	
	@Override
	public void setPreferredSpeed(double v_pref) {
		this.targetSpeed = v_pref;		
	}
	
	@Override
	public double getCurrentSpeed() {
		return currentSpeed;
	}

	/**
	 * Trigger the callback function, if there is a valid callback 
	 * 
	 * @param position
	 */
	private void locationReached(PositionVector position) {
		if(locationCallbacks.containsKey(position)) {
			locationCallbacks.get(position).reachedLocation();
		}
	}
	
	@Override
	public void setTargetLocation(PositionVector target,
			ReachedLocationCallback reachedLocationCallback) {
		route.clear();
		route.add(target);
		if(reachedLocationCallback != null)
			locationCallbacks.put(target, reachedLocationCallback);
		
	}

	@Override
	public void setTargetLocationRoute(LinkedList<PositionVector> route,
			ReachedLocationCallback reachedLocationCallback) {
		this.route.clear();
		this.route.addAll(route);
		if(reachedLocationCallback != null)
			locationCallbacks.put(route.getLast(), reachedLocationCallback);
	}

	@Override
	public void addTargetLocation(PositionVector target,
			ReachedLocationCallback reachedLocationCallback) {
		route.add(target);
		if(reachedLocationCallback != null)
			locationCallbacks.put(target, reachedLocationCallback);
	}

	@Override
	public LinkedList<PositionVector> getTargetLocations() {
		
		LinkedList<PositionVector> copy = new LinkedList<>();
		for (PositionVector pv : route) {
			copy.add(pv.clone());		
		}
		
		return copy;
	}

	@Override
	public void removeTargetLocations() {
		route.clear();
		locationCallbacks.clear();
	}	

	@Override
	public double horizontalMinVelocity() {
		return Math.sqrt(2 * hoverThrustRequired() * Math.tan(Math.toRadians(0.25)) / bodyDrag(Math.toRadians(0.25), new PositionVector(1,0,0)));
	}

	@Override
	public double estimatePowerConsumption(double velocity) {
		
		if(velocity == 0) {
			return motor.estimatePowerConsumptionWatt(hoverThrustRequired());
		}
		else if(velocity > horizontalMaxVelocity()) {
			return -1;
		}
		else if(velocity < horizontalMinVelocity()) {
			return -1;
		}
		else {			
			double estimateAngle = estimatePitchAngleForVelocity(velocity);
			double estimatedDrag = forwardDrag(velocity, estimateAngle);			
			
			double requiredThrust = Math.sqrt(hoverThrustRequired() * hoverThrustRequired() + estimatedDrag * estimatedDrag);			
			double wattage = motor.estimatePowerConsumptionWatt(requiredThrust);
						
			return wattage;
		}
		
		
	}
	
	public double estimateRequiredThrust(double velocity) {
		if(velocity == 0) {
			return motor.estimatePowerConsumptionWatt(hoverThrustRequired());
		}
		else if(velocity > horizontalMaxVelocity()) {
			return -1;
		}
		else if(velocity < horizontalMinVelocity()) {
			return -1;
		}
		else {
			double estimateAngle = estimatePitchAngleForVelocity(velocity);
			double estimatedDrag = forwardDrag(velocity, estimateAngle);					
			double requiredThrust = Math.sqrt(hoverThrustRequired() * hoverThrustRequired() + estimatedDrag * estimatedDrag);
			return requiredThrust;
		}
	}
	
	/**
	 * Estimate the pitch angle (angle of attack) required to get the target velocity. 
	 * Angle precision is 1/4 degree. 
	 * 
	 * @param velocity
	 * @return
	 */
	private double estimatePitchAngleForVelocity(double velocity) {
		
		int low = 0;
		int high = Integer.MAX_VALUE;
		
		double vsquared = (velocity * velocity);
		
		for(int i = 0; i <= ((int) Math.toDegrees(maxPitchAngle)); i++) {
			
			double v2 = 2 * hoverThrustRequired() * Math.tan(Math.toRadians(i)) / bodyDrag(Math.toRadians(i), new PositionVector(1,0,0));
						
			if(v2 > vsquared && i < high) {				
				high = i;
			}
			else if(v2 < vsquared && i >= low) {
				low = i;				
			}
			else if(v2 == vsquared ) {
				return Math.toRadians(i);
			}
			
		}
		
		if(high < Integer.MAX_VALUE) {
			double lo = low;
			double hi = high;
			
			double nearest = -1;
			double nearestDiff = Double.MAX_VALUE;
			
			double step = (hi - lo) / 4;
			
			for(int i = 0; i < 4; i++) {
			
				double d = lo + i * step;
				
				double v2 = 2 * hoverThrustRequired() * Math.tan(Math.toRadians(d)) / bodyDrag(Math.toRadians(d), new PositionVector(1,0,0));
				
				double diff = Math.abs(((velocity * velocity) - v2));
				
				if(diff < nearestDiff || (lo == 0 && i == 1)) {
					nearestDiff = diff;
					nearest = d;
				}			
			}
				
			return Math.toRadians(nearest);
		}
			
		return maxPitchAngle;	
	}	
	
	/**
	 * Factory for this movement model
	 * 
	 * @author Julian Zobel
	 * @version 1.0, 14.01.2020
	 */
	public static class Factory implements AerialMovementModelFactory {
		public UAVMovementModel createComponent(UAVTopologyComponent topologyComponent) {
			return new MulticopterMovement(topologyComponent);
		}		
	}
}
